/* Sources:
**   http://www.nongnu.org/avr-libc/user-manual/modules.html
** FUSE Calculator:
**   http://www.engbedded.com/cgi-bin/fc.cgi/
**
** FUSEs:  H:0xC8  L:0xEF
**   avrdude -c usbasp -p atmega32 -u -U hfuse:w:0xC8:m -U lfuse:w:0xEF:m
** Para leer los FUSEs:
**   avrdude -c usbasp -p atmega32 -U hfuse:r:hfuse.txt:m -U lfuse:r:lfuse.txt:m
**
** Grabar programa con una de las siguiente lineas:
**   avrdude -c usbasp -p atmega32 -U flash:w:Firmware.elf.hex
**   avrdude -c usbasp -p atmega32 -U flash:w:Firmware.elf.hex -U eeprom:w:Firmware.elf.eep.hex
*/

#define F_CPU 16000000UL

#include <avr/io.h>
#include <avr/interrupt.h>      // required by usbdrv.h
#include <avr/pgmspace.h>   	// required by usbdrv.h
#include <avr/wdt.h>            // WatchDog
#include <util/delay.h>			// _delay_ms()

#include "usbdrv/usbconfig.h"
#include "usbdrv/usbdrv.h"		// V-USB driver
//#include "usbdrv/oddebug.h"	// V-USB driver DEBUG (no lo usamos)


//=============================================================================
// USB, pines y mascaras
#define USB_MASK ( BIT(USB_CFG_DPLUS_BIT) | BIT(USB_CFG_DMINUS_BIT) )

//=============================================================================
// Constantes para la comunicacion con el software.
#define RQ_CUSTOM_ECHO              1
#define RQ_CUSTOM_SET_LEDS_STATUS   2
#define RQ_CUSTOM_GET_LEDS_STATUS   3

#define USB_HID_CUSTOM_RQ___ECHO                1
#define USB_HID_CUSTOM_RQ___LEDS_STATUS         2


//=============================================================================
// LEDs
#define LED_PORT_DDR   DDRC
#define LED_PORT_WRITE PORTC
#define LED_PORT_READ  PINC

//=============================================================================
// MACROS genericos
#define IO_CONCAT(a, b)  a ## b
#define IO_INPORT(name)  IO_CONCAT(PIN, name)
#define IO_OUTPORT(name) IO_CONCAT(PORT, name)
#define IO_DDRPORT(name) IO_CONCAT(DDR, name)
#define BIT(x) (1 << (x))  // BIT position
#define SBI(p,b) (p) |= BIT(b)
#define CBI(p,b) (p) &= ~BIT(b)
#define IS_SET(p,b) p & BIT(b)
#define IS_NOT_SET(p,b) ~p & BIT(b)
#define XOR(a,b) a^b
#ifndef NULL
 #define NULL    ((void *)0)
#endif
#define NOP() asm volatile("nop\n\t"::)


//
// El siguiente descriptor solo cumple la funcíon de silencia el driver de windows.
// Cuando windows pide por este descriptor, pensara que es un dispositivos HID,
//    y asignara un driver automaticamente, aun con este driver, libUSB puede enviar
//    todas las transacciones que necesitemos.
// Este descriptor debe estar configurado en el archivo usbconfig.h
//
//#define USB_CFG_DEVICE_CLASS          0
//#define USB_CFG_INTERFACE_CLASS     3
//#define USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH    22
PROGMEM char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] = {
    0x06, 0x00, 0xff,              // USAGE_PAGE (Generic Desktop)
    0x09, 0x01,                    // USAGE (Vendor Usage 1)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)
    0xc0                           // END_COLLECTION
};

static uchar reportBuffer[4];

static void softwareInit(void);
static void hardwareInit(void);

int main(void)
{
    uchar i;

    wdt_enable(WDTO_1S);
    // Si no usas WatchDog, siempre desactivalo al inicio
    // En algunos dispositivos nuevos, el estado del WD es preservado
    //   luego de un reseteo.

    // Tus funciones de inicializacion:
    hardwareInit();
	softwareInit();
    usbInit();


    // Forzar re-enumeración en caso de reseteo del dispositivo
    // Solo mientras las interrupciones se encuentran desactivadas
    usbDeviceDisconnect();
    i = 0;
    while(--i){     // Falsa desconexion de USB por menos de 250 ms
        wdt_reset();
        _delay_ms(1);
    }
    usbDeviceConnect();
    sei();			// Activar interrupciones globales


    // Loop principal
    while(1)
    {
        // Tus funciones Poll y las de tus librerias:
		wdt_reset();
		usbPoll();
    }


    return 0;
}



static void softwareInit(void)
{
	// Inicializar memorias y buffers
}

static void hardwareInit(void)
{
    // Inicializar puertos y modulos internos

    LED_PORT_DDR = 0xFF;
    LED_PORT_WRITE = 0x0F;

    // Configurar todo el puerto D como salida excepto los pines del USB
	DDRD = ~(USB_MASK);
	// No poner pullups en los pines del USB
	PORTD = ~(USB_MASK);

}

usbMsgLen_t usbFunctionSetup(uchar setupData[8])
{
    //
    // Al utilizar un descriptor HID empezamos a tener que respetar ciertos
    //    parametros que define el estandard, sin embargo mal usando el estandard HID
    //    podemos transmitir conjuntos de bytes arbitrarios.
    // Esto se hace con USBRQ_HID_GET_REPORT y USBRQ_HID_SET_REPORT
    //

	// Estructura de transaccion
    usbRequest_t  *rq = (void *)setupData;

    switch( rq->bmRequestType & USBRQ_TYPE_MASK ){
        case USBRQ_TYPE_VENDOR:
            //
            // Para peticiones de tipo Vendor mantenemos las mismas transacciones que
            //    para el ejemplo anterior para mantener compatibilidad con los software anteriores.
            //
            if(rq->bRequest == RQ_CUSTOM_ECHO)
            {
                // 1º Copiamos los datos:
                reportBuffer[0] = rq->wValue.bytes[0];
                reportBuffer[1] = rq->wValue.bytes[1];
                reportBuffer[2] = rq->wIndex.bytes[0];
                reportBuffer[3] = rq->wIndex.bytes[1];
                // 2º El puntero usbMsgPtr debe apuntar al inicio del buffer
                usbMsgPtr = reportBuffer;
                // 3º retornamos la cantidad de bytes. Menos de 8 bytes.
                // Si se excede de 8 bytes, la transacción automaticamente se divide en varios paquetes
                return 4;
            }
            else if(rq->bRequest == RQ_CUSTOM_SET_LEDS_STATUS)
            {
                LED_PORT_WRITE = rq->wValue.bytes[0];
            }
            else if(rq->bRequest == RQ_CUSTOM_GET_LEDS_STATUS)
            {
                reportBuffer[0] = LED_PORT_READ;
                usbMsgPtr = reportBuffer;
                return 1;
            }

            break;

        case USBRQ_TYPE_CLASS:

            switch(rq->bRequest){
                case USBRQ_HID_GET_REPORT:
                    //
                    // Por estandard, el parametro wValue recibido contiene las siguientes asignaciones:
                    //    wValue: ReportType (highbyte), ReportID (lowbyte)
                    // Lo cual no necesitamos respetar.
                    // En nuestro ejemplo usaremos el lowbyte de wIndex como identificador del reporte.
                    // Un reporte en este caso es solo una transferencia de datos.
                    //
                    switch( rq->wIndex.bytes[0] ){
						case USB_HID_CUSTOM_RQ___LEDS_STATUS:
                            reportBuffer[0]=LED_PORT_READ;
                            usbMsgPtr = reportBuffer;
							return 1;

						case USB_HID_CUSTOM_RQ___ECHO:
                            // 1º Copiamos los datos:
                            reportBuffer[0] = rq->wValue.bytes[0];
                            reportBuffer[1] = rq->wValue.bytes[1];
                            reportBuffer[2] = rq->wIndex.bytes[0];
                            reportBuffer[3] = rq->wIndex.bytes[1];
                            // 2º El puntero usbMsgPtr debe apuntar al inicio del buffer
                            usbMsgPtr = reportBuffer;
                            // 3º retornamos la cantidad de bytes. Menos de 8 bytes.
                            // Si se excede de 8 bytes, la transacción automaticamente se divide en varios paquetes
                            return 4;
                    }

                case USBRQ_HID_SET_REPORT:
                    switch( rq->wIndex.bytes[0] ){
                        case USB_HID_CUSTOM_RQ___LEDS_STATUS:
                            //
                            // Usamos el segundo byte de wIndex como selector de modo.
                            //
							switch( rq->wIndex.bytes[1] ){
                                case 1:
                                    LED_PORT_WRITE = 0x0f;
                                    break;
                                case 2:
                                    LED_PORT_WRITE = 0x55;
                                    break;
                                case 3:
                                    LED_PORT_WRITE = 0x01;
                                    break;
                                case 4:
                                    LED_PORT_WRITE = rq->wValue.bytes[0];
                                    break;
                                default:
                                    break;
								}
                            return 0;
                    }


            }
    }
    //
    // Si quisieramos implementar un mouse o un teclado, deberemos cambiar el
    //    descriptor e implementar las respectivas en la seccion Class
    //
    return 0;

}
