/*
** libusb API:
**    http://libusb.sourceforge.net/api-1.0/
**    http://libusb.sourceforge.net/api-1.0/group__syncio.html
**    http://libusb.sourceforge.net/api-1.0/group__dev.html#g7df95821d20d27b5597f1d783749d6a4
**
*/

#include <stdio.h>
#include <string.h>
#include "../Firmware/usbdrv/usbconfig.h"

#include <usb.h>

//
// Note aqui que ya no usamos USB_TYPE_VENDOR si no USB_TYPE_CLASS
//
#define USBRQ_HID_GET_REPORT    0x01
#define USBRQ_HID_SET_REPORT    0x09
#define CTRL_IN  (USB_TYPE_CLASS | USB_RECIP_DEVICE | USB_ENDPOINT_IN)
#define CTRL_OUT (USB_TYPE_CLASS | USB_RECIP_DEVICE | USB_ENDPOINT_OUT)

//=============================================================================
// Constantes para la comunicacion con el Firmware.
#define RQ_CUSTOM_ECHO 1
#define RQ_CUSTOM_SET_LEDS_STATUS 2
#define RQ_CUSTOM_GET_LEDS_STATUS 3

#define USB_HID_CUSTOM_RQ___ECHO                1
#define USB_HID_CUSTOM_RQ___LEDS_STATUS         2


usb_dev_handle* usb_find_and_open(int vendorID, char *vendor, int productID, char *product);

int main()
{
    //
    // Utilizando las definiciones en usbconfig.h definimos las variables respectivas
    //    y buscamos las mismas a traves de la lista de dispositivos.
    //
    const unsigned char rawVid[2] = {USB_CFG_VENDOR_ID};
    const unsigned char rawPid[2] = {USB_CFG_DEVICE_ID};
    char vendor[] = {USB_CFG_VENDOR_NAME, 0};
    char product[] = {USB_CFG_DEVICE_NAME, 0};
    int vendorID, productID;
    usb_dev_handle* usbdev;
    char buffer[8+1]={0};
    int rc;
    int wLength, wValue, wIndex;

    vendorID = rawVid[1] * 256 + rawVid[0];
    productID = rawPid[1] * 256 + rawPid[0];

    printf("Buscando: VID=%x, PID=%x, vendor=%s, product=%s\n",vendorID,productID,vendor,product);

    //
    // La libreria libusb debe ser siempre inicializada al inicio.
    // En el API de libusb-0.1 esta funcion no recibe ni devuelve ningun parametro.
    //
	usb_init();

    //
    // Implementamos una funcion que nos permita buscar el dipsositivo entre todos los conectados.
    //
    usbdev = usb_find_and_open(vendorID, vendor, productID, product);
    if (usbdev == NULL)
    {
	    fprintf(stderr, "failed to find device\n");
        return 0;
    }

    printf("Encontrado.\n");

    //
    // Aqui realizamos las transacciones que necesitemos y posteriormente,
    //    una vez finalizado cerramos la conexion con el dispositivo.
    //

    //
    // Las transferencias que se realizan, se hacen siempre como transacciones de control,
    //    de esta forma el dispositivo no necesita implementar el protocolo completo.
    //    int usb_control_msg( device_handler, request_type, request, wValue, wIndex, data, wLength, timeout )
    //

    //
    // En la siguiente transferencia enviamos 4 bytes y recibimos 4 bytes en un buffer de 8 bytes.
    //
    wLength = 8;
    wIndex = 0x2000 | USB_HID_CUSTOM_RQ___ECHO;
    wValue = 0x3106;
    rc = usb_control_msg( usbdev, CTRL_IN, USBRQ_HID_GET_REPORT, wValue, wIndex, buffer, wLength, 1000 );
    if ( rc < 0){
        fprintf(stderr, "failed to transfer a few bytes\n");
        usb_close(usbdev);
        return 0;
    }
    if ( !(rc==4 && buffer[0] == 0x06 && buffer[1] == 0x31 && buffer[2] == USB_HID_CUSTOM_RQ___ECHO && buffer[3] == 0x20) ){
	    fprintf(stderr, "failed to transfer safelly a few bytes\n");
        usb_close(usbdev);
        return 0;
    }
    printf("Realizado USB_HID_CUSTOM_RQ___ECHO - correctamente.\n");


    // ====================
    //
    // En la siguiente transferencie recibimos 1 byte en el buffer de 8.
    //
    wLength = 8;
    wIndex = USB_HID_CUSTOM_RQ___LEDS_STATUS;
    wValue = 0;
    rc = usb_control_msg( usbdev, CTRL_IN, USBRQ_HID_GET_REPORT, wValue, wIndex, buffer, wLength, 1000 );
    if ( rc < 0){
        fprintf(stderr, "failed to transfer a few bytes\n");
        usb_close(usbdev);
        return 0;
    }
    if ( rc!=1 ){
	    fprintf(stderr, "failed to transfer safelly a few bytes\n");
        usb_close(usbdev);
        return 0;
    }
    printf("Leido USB_HID_CUSTOM_RQ___LEDS_STATUS - correctamente. Valor 0x%02x\n",buffer[0]);

    //
    // Aqui enviamos un valor que sera usado en un puerto de LEDs.
    // y no esperamos respusta.
    //
    wLength = 0;
    wIndex = (4 <<8) | USB_HID_CUSTOM_RQ___LEDS_STATUS;
    wValue = 0x55;
    rc = usb_control_msg( usbdev, CTRL_OUT, USBRQ_HID_SET_REPORT, wValue, wIndex, 0, wLength, 1000 );
    if ( rc < 0){
        fprintf(stderr, "failed to transfer a few bytes\n");
        usb_close(usbdev);
        return 0;
    }
    if ( rc!=0 ){
        fprintf(stderr, "failed to transfer safelly a few bytes  %d\n",rc);
        usb_close(usbdev);
        return 0;
    }
    printf("Escrito USB_HID_CUSTOM_RQ___LEDS_STATUS modo 4 - correctamente. Valor 0x%02x\n",wValue);

    //
    // Finalizado el programa, liberamos la memoria.
    //
    usb_close(usbdev);
    return 0;
}


usb_dev_handle* usb_find_and_open(int vendorID, char *vendor, int productID, char *product)
{
    usb_dev_handle *handle;
    struct usb_bus      *bus;
    struct usb_device   *dev;
    char buffer[60];
    int len=0;

    //
    // Lo primero es obtener una lista de los dispositivos usb conectados.
    // Eso incluye los HUBs y todo el arbol de dispositivos.
    // LibUSB-0.1 nos provee de dos elementos iteradores,
    //    los buses y los dispositivos.
    //

    usb_find_busses();
    usb_find_devices();
    for(bus = usb_get_busses(); bus; bus = bus->next){
        for(dev = bus->devices; dev; dev = dev->next){

        //
        // Para obtener los parámetros del dispositivo debemos pedir su descriptor
        //   de dispositivo en dev->descriptor. Los detalles de sus miembros se pueden leer del mismo header
        //   usb.h y debemos saber que sus nombres estan definidos por el estandard y no por el API.
        //

        //
        // Como primer paso a identificar nuestro dispositivo buscamos cada uno
        //    que tenga el par VID/PID del nuestro. Pero puede haber mas de un
        //    dispositivo conectado con dicho par.
        //
        if ( ! (vendorID==0
             || dev->descriptor.idVendor == vendorID) )
            continue;

        if ( ! (productID == 0
             || dev->descriptor.idProduct == productID) )
            continue;

        printf("--> VID=%x, PID=%x\n",vendorID,productID);

        //
        // Luego necesitaremos entrar en detalles, seleccionar el dispositivo correcto
        //    para ello chequeamos el nombre este pidiendo por los descriptor de String.
        //    En tales, se puede encontrar el nombre textual del manufacturador, del dispositivo
        //    Y mas importante, del numero de serie.
        //
        handle = usb_open(dev);
        if (!handle){
            fprintf(stderr, "failed to open device\n");
            continue;
            }


        len = buffer[0] = 0;
        if(dev->descriptor.iManufacturer > 0){
            len = usb_get_string_simple(handle, dev->descriptor.iManufacturer, buffer, sizeof(buffer));
            if(len < 0)
                fprintf(stderr, "cannot query manufacturer for VID=0x%04x PID=0x%04x\n", vendorID, productID);
            else if( strcmp(buffer, vendor) !=0 )
                {   usb_close(handle);
                    continue;
                }
        }
        else { usb_close(handle);
                continue;
        }

        printf("--> vendor=%s\n",buffer);

        len = buffer[0] = 0;
        if(dev->descriptor.iProduct > 0){
            len = usb_get_string_simple(handle, dev->descriptor.iProduct, buffer, sizeof(buffer));
            if(len < 0)
                fprintf(stderr, "cannot query product for VID=0x%04x PID=0x%04x\n", vendorID, productID);
            else if( strcmp(buffer, product) !=0 )
                {   usb_close(handle);
                    continue;
                }
        }
        else { usb_close(handle);
                continue;
        }

        printf("--> product=%s\n",buffer);

        //
        // Para nuestro caso no incluiremos el numero de Serie y se deja como ejercicio para el lector
        //

        //
        // Finalmente, para este punto, la variable handle ya contiene el dispositivo,
        //    y es la variable que debemos retornar.
        //
        return handle;
    }}

    //
    // Si el dispositivo no se ha encontrado, retornamos NULL.
    //
    return NULL;
}

