/*
** libusb API:
**    http://libusb.sourceforge.net/api-1.0/
**    http://libusb.sourceforge.net/api-1.0/group__syncio.html
**    http://libusb.sourceforge.net/api-1.0/group__dev.html#g7df95821d20d27b5597f1d783749d6a4
**
*/

#include <stdio.h>
#include <string.h>
#include "../Firmware/usbdrv/usbconfig.h"

#include <libusb-1.0/libusb.h>

//
// Note aqui que ya no usamos LIBUSB_REQUEST_TYPE_VENDOR si no LIBUSB_REQUEST_TYPE_CLASS
//
#define USBRQ_HID_GET_REPORT    0x01
#define USBRQ_HID_SET_REPORT    0x09
#define CTRL_IN  (LIBUSB_REQUEST_TYPE_CLASS | LIBUSB_RECIPIENT_DEVICE | LIBUSB_ENDPOINT_IN)
#define CTRL_OUT (LIBUSB_REQUEST_TYPE_CLASS | LIBUSB_RECIPIENT_DEVICE | LIBUSB_ENDPOINT_OUT)

//=============================================================================
// Constantes para la comunicacion con el Firmware.
#define RQ_CUSTOM_ECHO 1
#define RQ_CUSTOM_SET_LEDS_STATUS 2
#define RQ_CUSTOM_GET_LEDS_STATUS 3

#define USB_HID_CUSTOM_RQ___ECHO                1
#define USB_HID_CUSTOM_RQ___LEDS_STATUS         2


libusb_device_handle* usb_find_and_open(int vendorID, char *vendor, int productID, char *product);

int main()
{
    //
    // Utilizando las definiciones en usbconfig.h definimos las variables respectivas
    //    y buscamos las mismas a traves de la lista de dispositivos.
    //
    const unsigned char rawVid[2] = {USB_CFG_VENDOR_ID};
    const unsigned char rawPid[2] = {USB_CFG_DEVICE_ID};
    char vendor[] = {USB_CFG_VENDOR_NAME, 0};
    char product[] = {USB_CFG_DEVICE_NAME, 0};
    int vendorID, productID;
    libusb_device_handle* usbdev;
    unsigned char buffer[8+1]={0};
    int rc;
    int wLength, wValue, wIndex;

    vendorID = rawVid[1] * 256 + rawVid[0];
    productID = rawPid[1] * 256 + rawPid[0];

    printf("Buscando: VID=%x, PID=%x, vendor=%s, product=%s\n",vendorID,productID,vendor,product);

    //
    // La libreria libusb debe ser siempre inicializada al inicio.
    // Corroborar que no halla fallado es muy importante, ya que esta puede fallar
    //   si por ejemplo, el software no tiene permisos para usar los dispositivos usb
    //   o si el driver/filtro de libusb no ha sido instalado correctamente.
    //
	if ( libusb_init(NULL) != 0 ) {
	    fprintf(stderr, "failed to init libusb\n");
        return 0;
	}

    //
    // Implementamos una funcion que nos permita buscar el dipsositivo entre todos los conectados.
    //
    usbdev = usb_find_and_open(vendorID, vendor, productID, product);
    if (usbdev == NULL)
    {
	    fprintf(stderr, "failed to find device\n");
        libusb_exit(NULL);
        return 0;
    }

    printf("Encontrado.\n");

    //
    // Previo a realizar las transacciones de I/O, debemos "reclamar la interfaz".
    // Esta es una operacion puramente logica, entre el sistema operativo y tu programa.
    //

    if ( libusb_claim_interface(usbdev, 0) !=0 ){
	    fprintf(stderr, "failed to claim interface\n");
        libusb_close(usbdev);
        libusb_exit(NULL);
        return 0;
    }

    //
    // Aqui realizamos las transacciones que necesitemos y posteriormente,
    //    una vez finalizado cerramos la conexion con el dispositivo,
    //    y llamamos a la funcion exit de la libreria libusb para que elimine
    //    las variables que halla creado.
    //

    //
    // Las transferencias que se realizan, se hacen siempre como transacciones de control,
    //    de esta forma el dispositivo no necesita implementar el protocolo completo.
    //    int usb_control_msg( device_handler, request_type, request, wValue, wIndex, data, wLength, timeout )
    //

    //
    // En la siguiente transferencia enviamos 4 bytes y recibimos 4 bytes en un buffer de 8 bytes.
    //
    wLength = 8;
    wIndex = 0x2000 | USB_HID_CUSTOM_RQ___ECHO;
    wValue = 0x3106;
    rc = libusb_control_transfer(usbdev, CTRL_IN, USBRQ_HID_GET_REPORT, wValue, wIndex, buffer, wLength, 1000 );
    if ( rc < 0){
        fprintf(stderr, "failed to transfer a few bytes\n");
        libusb_release_interface(usbdev, 0);
        libusb_close(usbdev);
        libusb_exit(NULL);
        return 0;
    }
    if ( !(rc==4 && buffer[0] == 0x06 && buffer[1] == 0x31 && buffer[2] == USB_HID_CUSTOM_RQ___ECHO && buffer[3] == 0x20) ){
	    fprintf(stderr, "failed to transfer safelly a few bytes\n");
        libusb_release_interface(usbdev, 0);
        libusb_close(usbdev);
        libusb_exit(NULL);
        return 0;
    }
    printf("Realizado USB_HID_CUSTOM_RQ___ECHO - correctamente.\n");

    //
    // En la siguiente transferencie recibimos 1 byte en el buffer de 8.
    //
    wLength = 8;
    wIndex = USB_HID_CUSTOM_RQ___LEDS_STATUS;
    wValue = 0;
    rc = libusb_control_transfer(usbdev, CTRL_IN, USBRQ_HID_GET_REPORT, wValue, wIndex, buffer, wLength, 1000 );
    if ( rc < 0){
        fprintf(stderr, "failed to transfer a few bytes\n");
        libusb_release_interface(usbdev, 0);
        libusb_close(usbdev);
        libusb_exit(NULL);
        return 0;
    }
    if ( rc!=1 ){
	    fprintf(stderr, "failed to transfer safelly a few bytes\n");
        libusb_release_interface(usbdev, 0);
        libusb_close(usbdev);
        libusb_exit(NULL);
        return 0;
    }
    printf("Leido USB_HID_CUSTOM_RQ___LEDS_STATUS - correctamente. Valor 0x%02x\n",buffer[0]);

    //
    // En la siguiente transferencia enviamos 1 byte.
    //
    wLength = 0;
    wIndex = (4 <<8) | USB_HID_CUSTOM_RQ___LEDS_STATUS;
    wValue = 0x55;
    rc = libusb_control_transfer(usbdev, CTRL_OUT, USBRQ_HID_SET_REPORT, wValue, wIndex, 0, wLength, 1000 );
    if ( rc < 0){
        fprintf(stderr, "failed to transfer a few bytes\n");
        libusb_release_interface(usbdev, 0);
        libusb_close(usbdev);
        libusb_exit(NULL);
        return 0;
    }
    if ( rc!=0 ){
        fprintf(stderr, "failed to transfer safelly a few bytes  %d\n",rc);
        libusb_release_interface(usbdev, 0);
        libusb_close(usbdev);
        libusb_exit(NULL);
        return 0;
    }
    printf("Escrito USB_HID_CUSTOM_RQ___LEDS_STATUS modo 4 - correctamente. Valor 0x%02x\n",wValue);

    //
    // Finalizado el programa, liberamos la memoria.
    //
    libusb_release_interface(usbdev, 0);
    libusb_close(usbdev);
    libusb_exit(NULL);
    return 0;
}


libusb_device_handle* usb_find_and_open(int vendorID, char *vendor, int productID, char *product)
{
    libusb_device_handle *handle;
    libusb_device **list;
    size_t i = 0;
    size_t cnt;
    int err = 0;
    char buffer[60];
    int len=0;

    //
    // Lo primero es obtener una lista de los dispositivos usb conectados.
    // Eso incluye los HUBs y todo el arbol de dispositivos.
    // LibUSB-1.0 nos provee de una sencilla lista en la cual deberemos iterar.
    //
    cnt = libusb_get_device_list(NULL, &list);
    if (cnt < 0){
        fprintf(stderr, "failed to get device list\n");
    }

    for (i = 0; i < cnt; i++)
    {
        libusb_device *dev = list[i];
        struct libusb_device_descriptor desc;

        //
        // Para obtener los parámetros del dispositivo debemos pedir su descriptor
        //   de dispositivo. Los detalles de sus miembros se pueden leer del mismo header
        //   libusb-1.0/libusb.h
        //

        int r = libusb_get_device_descriptor(dev, &desc);
        if (r < 0) {
            fprintf(stderr, "failed to get device descriptor\n");
            continue;
        }

        //
        // Como primer paso a identificar nuestro dispositivo buscamos cada uno
        //    que tenga el par VID/PID del nuestro. Pero puede haber mas de un
        //    dispositivo conectado con dicho par.
        //
        if ( ! (vendorID==0
             || desc.idVendor == vendorID) )
            continue;

        if ( ! (productID == 0
             || desc.idProduct == productID) )
            continue;

        printf("--> VID=%x, PID=%x\n",vendorID,productID);

        //
        // Luego necesitaremos entrar en detalles, seleccionar el dispositivo correcto
        //    para ello chequeamos el nombre este pidiendo por los descriptor de String.
        //    En tales, se puede encontrar el nombre textual del manufacturador, del dispositivo
        //    Y mas importante, del numero de serie.
        //
        err = libusb_open(dev, &handle);
        if (err){
            fprintf(stderr, "failed to open device\n");
            continue;
            }


        len = buffer[0] = 0;
        if(desc.iManufacturer > 0){
            len = libusb_get_string_descriptor_ascii(handle, desc.iManufacturer, (unsigned char*)buffer, sizeof(buffer));
            if(len < 0)
                fprintf(stderr, "cannot query manufacturer for VID=0x%04x PID=0x%04x\n", vendorID, productID);
            else if( strcmp(buffer, vendor) !=0 )
                {   libusb_close(handle);
                    continue;
                }
        }
        else { libusb_close(handle);
                continue;
        }

        printf("--> vendor=%s\n",buffer);

        len = buffer[0] = 0;
        if(desc.iProduct > 0){
            len = libusb_get_string_descriptor_ascii(handle, desc.iProduct, (unsigned char*)buffer, sizeof(buffer));
            if(len < 0)
                fprintf(stderr, "cannot query product for VID=0x%04x PID=0x%04x\n", vendorID, productID);
            else if( strcmp(buffer, product) !=0 )
                {   libusb_close(handle);
                    continue;
                }
        }
        else { libusb_close(handle);
                continue;
        }

        printf("--> product=%s\n",buffer);

        //
        // Para nuestro caso no incluiremos el numero de Serie y se deja como ejercicio para el lector
        //

        //
        // Por ultimo, liberar de la memoria la lista de dispositivos.
        // Finalmente, para este punto, la variable handle ya contiene el dispositivo,
        //    y es la variable que debemos retornar.
        //
        libusb_free_device_list(list, 1);
        return handle;
    }

    //
    // Si el dispositivo no se ha encontrado, retornamos NULL.
    //
    libusb_free_device_list(list, 1);
    return NULL;
}

