/* Sources:
**   http://www.nongnu.org/avr-libc/user-manual/modules.html
** FUSE Calculator:
**   http://www.engbedded.com/cgi-bin/fc.cgi/
**
** FUSEs:  H:0xC8  L:0xEF
**   avrdude -c usbasp -p atmega32 -u -U hfuse:w:0xC8:m -U lfuse:w:0xEF:m
** Para leer los FUSEs:
**   avrdude -c usbasp -p atmega32 -U hfuse:r:hfuse.txt:m -U lfuse:r:lfuse.txt:m
**
** Grabar programa con una de las siguiente lineas:
**   avrdude -c usbasp -p atmega32 -U flash:w:Firmware.elf.hex
**   avrdude -c usbasp -p atmega32 -U flash:w:Firmware.elf.hex -U eeprom:w:Firmware.elf.eep.hex
*/

#define F_CPU 16000000UL

#include <avr/io.h>
#include <avr/interrupt.h>      // required by usbdrv.h
#include <avr/pgmspace.h>   	// required by usbdrv.h
#include <avr/wdt.h>            // WatchDog
#include <util/delay.h>			// _delay_ms()

#include "usbdrv/usbconfig.h"
#include "usbdrv/usbdrv.h"		// V-USB driver
//#include "usbdrv/oddebug.h"	// V-USB driver DEBUG (no lo usamos)


//=============================================================================
// USB, pines y mascaras
#define USB_MASK ( BIT(USB_CFG_DPLUS_BIT) | BIT(USB_CFG_DMINUS_BIT) )

//=============================================================================
// Constantes para la comunicacion con el software.
#define RQ_CUSTOM_ECHO 1
#define RQ_CUSTOM_SET_LEDS_STATUS 2
#define RQ_CUSTOM_GET_LEDS_STATUS 3

//=============================================================================
// LEDs
#define LED_PORT_DDR   DDRC
#define LED_PORT_WRITE PORTC
#define LED_PORT_READ  PINC

//=============================================================================
// MACROS genericos
#define IO_CONCAT(a, b)  a ## b
#define IO_INPORT(name)  IO_CONCAT(PIN, name)
#define IO_OUTPORT(name) IO_CONCAT(PORT, name)
#define IO_DDRPORT(name) IO_CONCAT(DDR, name)
#define BIT(x) (1 << (x))  // BIT position
#define SBI(p,b) (p) |= BIT(b)
#define CBI(p,b) (p) &= ~BIT(b)
#define IS_SET(p,b) p & BIT(b)
#define IS_NOT_SET(p,b) ~p & BIT(b)
#define XOR(a,b) a^b
#ifndef NULL
 #define NULL    ((void *)0)
#endif
#define NOP() asm volatile("nop\n\t"::)


static uchar reportBuffer[4];

static void softwareInit(void);
static void hardwareInit(void);
// Las funciones static copiarán siempre su contenido a donde son llamadas
// Sin este modificador, será el compilador quien decida optimizar
//   de esta forma. Normalmente toda funcion llamada solo desde 1 lugar
//   se compila cono static.
// Esto es bueno para evitar las instrucciones CALL, PUSH y PULL en exceso.


int main(void)
{
    uchar i;

    wdt_enable(WDTO_1S);
    // Si no usas WatchDog, siempre desactivalo al inicio
    // En algunos dispositivos nuevos, el estado del WD es preservado
    //   luego de un reseteo.

    // Tus funciones de inicializacion:
    hardwareInit();
	softwareInit();
    usbInit();


    // Forzar re-enumeración en caso de reseteo del dispositivo
    // Solo mientras las interrupciones se encuentran desactivadas
    usbDeviceDisconnect();
    i = 0;
    while(--i){     // Falsa desconexion de USB por menos de 250 ms
        wdt_reset();
        _delay_ms(1);
    }
    usbDeviceConnect();
    sei();			// Activar interrupciones globales


    // Loop principal
    while(1)
    {
        // Tus funciones Poll y las de tus librerias:
		wdt_reset();
		usbPoll();
    }


    return 0;
}



static void softwareInit(void)
{
	// Inicializar memorias y buffers
}

static void hardwareInit(void)
{
    // Inicializar puertos y modulos internos

    LED_PORT_DDR = 0xFF;
    LED_PORT_WRITE = 0x0F;

    // Configurar todo el puerto D como salida excepto los pines del USB
	DDRD = ~(USB_MASK);
	// No poner pullups en los pines del USB
	PORTD = ~(USB_MASK);

}

// Cuando el Host envia un paquete de control al endpoint 0 de
// tipo "Vendor" o "Class" se llama a la funcion usbFunctionSetup().
// Para enviar datos al host se debe setear el usbMsgPtr al inicio de datos y retornar el largo.
usbMsgLen_t usbFunctionSetup(uchar setupData[8])
{
	// Estructura de transaccion
    usbRequest_t  *rq = (void *)setupData;

    if(rq->bRequest == RQ_CUSTOM_ECHO)
    {
        // La transaccion provee de un numero de campos de libre uso de 16 bits
        //    estos son:  wValue, wIndex, wLength.  Con los cuales se puede enviar
        //    informacion desde el host al device en un solo paquete.
        // El dispositivo podrá retornar hasta 8 bytes de datos por paquete.
        // Para mas de 8 bytes se usa una función especial que veremos en otro ejemplo.

        // En este ejemplo, copiamos las variables que provienen del host
        //    y las reenviamos en el paquete de datos, de modo que en el software
        //    verificaremos que los datos se han recibido correctamente.

        // 1º Copiamos los datos:
        reportBuffer[0] = rq->wValue.bytes[0];
        reportBuffer[1] = rq->wValue.bytes[1];
        reportBuffer[2] = rq->wIndex.bytes[0];
        reportBuffer[3] = rq->wIndex.bytes[1];
        // 2º El puntero usbMsgPtr debe apuntar al inicio del buffer
        usbMsgPtr = reportBuffer;
        // 3º retornamos la cantidad de bytes a transmitir.
        // Si se excede de 8 bytes, la transacción automaticamente se divide en varios paquetes
        return 4;
    }
    else if(rq->bRequest == RQ_CUSTOM_SET_LEDS_STATUS)
    {
        LED_PORT_WRITE = rq->wValue.bytes[0];
    }
    else if(rq->bRequest == RQ_CUSTOM_GET_LEDS_STATUS)
    {
        reportBuffer[0] = LED_PORT_READ;
        usbMsgPtr = reportBuffer;
        return 1;
    }
    // Retorno por defecto para peticiones no implementadas:
    // No retornar datos al Host.
    return 0;

    // Algunos tipos de transacciones no necesitan devolver ninguna
    //    confirmación ni datos, en este caso se retorna 0.
    // Si fuese necesario devolver datos al host, ya se para
    //    verificar recepción, o transmitir los datos que el Host pide,
    //    se debe siempre asignar el puntero usbMsgPtr al inicio de un buffer
    //    determinado como se ve anteriormente.
    //    Y luego retornar el tamaño del buffer en bytes.
    //    Si el tamaño es de menos de 8 bytes, los datos se transmitiran
    //    en una unica transmición, si fuese mayor, los datos se transmitiran
    //    en multiples paquetes y deberemos implementar una función llamada
    //    usbFunctionRead, cuando el dispositivo envia al host, o usbFunctionWrite
    //    cuando el Host envía al dispositivo. Ambas se ven en otro ejemplo.
}
