/*
** libusb API:
**    http://libusb.sourceforge.net/api-1.0/
**    http://libusb.sourceforge.net/api-1.0/group__syncio.html
**    http://libusb.sourceforge.net/api-1.0/group__dev.html#g7df95821d20d27b5597f1d783749d6a4
**
*/

#include <stdio.h>
#include <stdlib.h>

//
// libusb 1.0 tiene un header diferente que la 0.1
//     y tambien es diferente la libreria con la que hay que linkear
//
#include <libusb-1.0/libusb.h>

int main()
{
    int rc;

    //
    // La funcion de inicializacion ahora retorna un valor que indica los posibles
    //     errores que pueden haber. Entre ellos que no este instalado el driver,
    //     que no tenga parmisos para acceder a los dispositivos usb, etc.
    //
	rc = libusb_init(NULL);
	if (rc == 0) {

        size_t i = 0;
        int err = 0;
        unsigned char buffer[250];
        int len=0;

        libusb_device_handle *handle;
        libusb_device **list;
        //
        // En libusb 1.0 iterar sobre todos los dispositivos es mas simple
        //    ya que provee de la siguiente funcion que retorna la lista completa.
        //    Ademas nos permite proteger el software ante eventuales errores.
        //
        ssize_t cnt = libusb_get_device_list(NULL, &list);
        if (cnt < 0)
            printf("Error(libusb_get_device_list)\n");

        for (i = 0; i < cnt; i++)
        {
            libusb_device *dev = list[i];
            struct libusb_device_descriptor desc;

            //
            // El descriptor es ahora una estructura que se debe mandar a pedir
            //     Este descriptor esta normalmente almacenado por el driver cuando
            //     el dispositivo se conecta al host, así que no involucra ninguna
            //     transacción.
            //
            int r = libusb_get_device_descriptor(dev, &desc);
            if (r < 0) {
                fprintf(stderr, "failed to get device descriptor");
                continue;
            }

            //
            // Nuevamente, en libusb 1.0 podemos proteger el software ante eventuales errores
            //
            err = libusb_open(dev, &handle);
            if (err){
                printf("Error(libusb_open)\n");
                continue;
                }

            printf("idVendor(VID): %x \t",desc.idVendor);
            printf("idProduct(PID): %x \n",desc.idProduct);

            //
            // libusb_get_string_descriptor_ascii es equivalente y toma los mismos parametros que
            //     usb_get_string_simple
            //
            //printf("iManufacturer: %x \t",desc.iManufacturer);
            if(desc.iManufacturer > 0){
                len = libusb_get_string_descriptor_ascii(handle, desc.iManufacturer, buffer, sizeof(buffer));
                buffer[len]='\0';
                printf("Manufacturer: %s\n", buffer);
            }

            //printf("iProduct: %x \t",desc.iProduct);
            if(desc.iProduct > 0){
                len = libusb_get_string_descriptor_ascii(handle, desc.iProduct, buffer, sizeof(buffer));
                buffer[len]='\0';
                printf("Product: %s\n", buffer);
            }

            //printf("iSerialNumber: %x \t",desc.iSerialNumber);
            if(desc.iSerialNumber > 0){
                len = libusb_get_string_descriptor_ascii(handle, desc.iSerialNumber, buffer, sizeof(buffer));
                buffer[len]='\0';
                printf("SerialNumber: %s\n", buffer);
            }

            printf("bLength: %x \t",desc.bLength);
            printf("bDeviceClass: %x \n",desc.bDeviceClass);
            //
            // Una ventaja adicional de libusb 1.0 es que esta muy bien documentado.
            // No solo pueden encontrarse las definiciones que podemos usar si no tambien explicaciones
            //     y otros detalles.
            //
            switch (desc.bDeviceClass){
                case LIBUSB_CLASS_PER_INTERFACE:
                    printf("DeviceClass: PER_INTERFACE\n");
                    break;
                case LIBUSB_CLASS_AUDIO:
                    printf("DeviceClass: UDI\n");
                    break;
                case LIBUSB_CLASS_COMM:
                    printf("DeviceClass: COMM\n");
                    break;
                case LIBUSB_CLASS_HID:
                    printf("DeviceClass: HID\n");
                    break;
                case LIBUSB_CLASS_PRINTER:
                    printf("DeviceClass: PRINTER\n");
                    break;
                case LIBUSB_CLASS_PTP:
                    printf("DeviceClass: PTP\n");
                    break;
                case LIBUSB_CLASS_MASS_STORAGE:
                    printf("DeviceClass: MASS_STORAGE\n");
                    break;
                case LIBUSB_CLASS_HUB:
                    printf("DeviceClass: HUB\n");
                    break;
                case LIBUSB_CLASS_DATA:
                    printf("DeviceClass: DATA\n");
                    break;
                case LIBUSB_CLASS_VENDOR_SPEC:
                    printf("DeviceClass: VENDOR_SPEC\n");
                    break;
                default:
                    printf("DeviceClass: ¿ERROR?\n");
                    break;
            }

            //printf("bDescriptorType: %x \t",desc.bDescriptorType);
            //printf("bDeviceSubClass: %x \t",desc.bDeviceSubClass);
            //printf("bcdUSB: %x \t",desc.bcdUSB);
            //printf("bDeviceProtocol: %x \t",desc.bDeviceProtocol);
            //printf("bMaxPacketSize0: %x \t",desc.bMaxPacketSize0);
            //printf("bcdDevice: %x \t",desc.bcdDevice);
            //printf("bNumConfigurations: %x \t",desc.bNumConfigurations);
            printf("\n");

            libusb_close(handle);
        }

        //
        // Al finalizar toda memoria ocupada debe ser liberada.
        //
        libusb_free_device_list(list, 1);

		libusb_exit(NULL);
	}

    return 0;
}
